from collections.abc import Iterator

import pytest
from graph_retriever.testing.adapter_tests import AdapterComplianceSuite
from langchain_astradb.utils.vector_store_codecs import _DefaultVSDocumentCodec
from langchain_core.documents import Document
from langchain_core.embeddings import Embeddings
from langchain_graph_retriever.adapters.astra import AstraAdapter, _QueryHelper

TEST_CODEC = _DefaultVSDocumentCodec("page_content", ignore_invalid_documents=True)


def test_create_ids_query_no_user() -> None:
    query_helper = _QueryHelper(TEST_CODEC, {})

    assert query_helper.create_ids_query(["1"]) == {"_id": "1"}

    query = query_helper.create_ids_query(["1", "2", "3"])
    assert query is not None
    assert query["_id"]["$in"] == ["1", "2", "3"]


def test_create_ids_query_user() -> None:
    query_helper = _QueryHelper(TEST_CODEC, {"answer": 42})

    assert query_helper.create_ids_query(["1"]) == {
        "$and": [
            {"_id": "1"},
            {"metadata.answer": 42},
        ]
    }

    query = query_helper.create_ids_query(["1", "2", "3"])
    assert query is not None
    assert query["$and"][0]["_id"]["$in"] == ["1", "2", "3"]


def test_create_metadata_query_no_user() -> None:
    query_helper = _QueryHelper(TEST_CODEC, {})

    assert query_helper.create_metadata_query({}) is None
    assert query_helper.create_metadata_query({"foo": []}) is None
    assert query_helper.create_metadata_query({"foo": [5]}) == {"metadata.foo": 5}

    query = query_helper.create_metadata_query({"foo": [5, 6]})
    assert query is not None
    assert sorted(query["metadata.foo"]["$in"]) == [5, 6]

    assert query_helper.create_metadata_query({"foo": [5], "bar": [7]}) == {
        "$or": [
            {"metadata.foo": 5},
            {"metadata.bar": 7},
        ],
    }

    query = query_helper.create_metadata_query({"foo": [5, 6], "bar": [7, 8]})
    assert query is not None
    assert sorted(query["$or"][0]["metadata.foo"]["$in"]) == [5, 6]
    assert sorted(query["$or"][1]["metadata.bar"]["$in"]) == [7, 8]

    query = query_helper.create_metadata_query({"foo": list(range(0, 200))})
    assert query is not None
    assert sorted(query["$or"][0]["metadata.foo"]["$in"]) == list(range(0, 100))
    assert sorted(query["$or"][1]["metadata.foo"]["$in"]) == list(range(100, 200))


def test_create_metadata_query_user() -> None:
    query_helper = _QueryHelper(TEST_CODEC, {"answer": 42})

    assert query_helper.create_metadata_query({}) is None
    assert query_helper.create_metadata_query({"foo": []}) is None
    assert query_helper.create_metadata_query({"foo": [5]}) == {
        "$and": [
            {"metadata.foo": 5},
            {"metadata.answer": 42},
        ],
    }

    query = query_helper.create_metadata_query({"foo": [5, 6]})
    assert query is not None
    assert query["$and"][0]["metadata.foo"]["$in"] == [5, 6]
    assert query["$and"][1] == {"metadata.answer": 42}

    assert query_helper.create_metadata_query({"foo": [5], "bar": [7]}) == {
        "$and": [
            {
                "$or": [
                    {"metadata.foo": 5},
                    {"metadata.bar": 7},
                ],
            },
            {"metadata.answer": 42},
        ],
    }

    query = query_helper.create_metadata_query({"foo": [5, 6], "bar": [7, 8]})
    assert query is not None
    assert query["$and"][0]["$or"][0]["metadata.foo"]["$in"] == [5, 6]
    assert query["$and"][0]["$or"][1]["metadata.bar"]["$in"] == [7, 8]
    assert query["$and"][1] == {"metadata.answer": 42}

    query = query_helper.create_metadata_query({"foo": list(range(0, 200))})
    assert query is not None
    assert query["$and"][0]["$or"][0]["metadata.foo"]["$in"] == list(range(0, 100))
    assert query["$and"][0]["$or"][1]["metadata.foo"]["$in"] == list(range(100, 200))
    assert query["$and"][1] == {"metadata.answer": 42}


class TestAstraAdapter(AdapterComplianceSuite):
    @pytest.fixture(scope="class")
    def adapter(
        self,
        enabled_stores: set[str],
        animal_embeddings: Embeddings,
        animal_docs: list[Document],
    ) -> Iterator["AstraAdapter"]:
        if "astra" not in enabled_stores:
            pytest.skip("Pass --stores=astra to test Astra")
            return

        import os

        from astrapy import AstraDBDatabaseAdmin
        from astrapy.authentication import StaticTokenProvider
        from dotenv import load_dotenv
        from langchain_astradb import AstraDBVectorStore
        from langchain_graph_retriever.adapters.astra import (
            AstraAdapter,
        )

        load_dotenv()

        token = StaticTokenProvider(os.environ["ASTRA_DB_APPLICATION_TOKEN"])
        keyspace = os.environ.get("ASTRA_DB_KEYSPACE", "default_keyspace")
        api_endpoint = os.environ["ASTRA_DB_API_ENDPOINT"]

        admin = AstraDBDatabaseAdmin(api_endpoint=api_endpoint, token=token)
        admin.create_keyspace(keyspace)

        store = AstraDBVectorStore(
            embedding=animal_embeddings,
            collection_name="animals",
            namespace=keyspace,
            token=token,
            api_endpoint=api_endpoint,
            pre_delete_collection=True,
        )
        store.add_documents(animal_docs)

        yield AstraAdapter(store)

        store.delete_collection()



from collections.abc import Sequence
from typing import Any

from langchain_core.documents import BaseDocumentTransformer, Document
from typing_extensions import override


class GLiNEREntityExtractor(BaseDocumentTransformer):
    """
    Add metadata to documents about named entities using `GLiNER`_.

    `GLiNER`_ is a Named Entity Recognition (NER) model capable of identifying any
    entity type using a bidirectional transformer encoder (BERT-like).

    Preliminaries
    -------------

    Install the ``gliner`` package.

    Note that ``bs4`` is also installed to support the WebBaseLoader in the example,
    but not needed by the GLiNEREntityExtractor itself.

    .. code-block:: bash

        pip install -q langchain_community bs4 gliner

    Example
    -------
    We load the ``state_of_the_union.txt`` file, chunk it, then for each chunk we
    add named entities to the metadata.

    .. code-block:: python

        from langchain_community.document_loaders import WebBaseLoader
        from langchain_community.document_transformers import GLiNEREntityExtractor
        from langchain_text_splitters import CharacterTextSplitter

        loader = WebBaseLoader(
            "https://raw.githubusercontent.com/hwchase17/chat-your-data/master/state_of_the_union.txt"
        )
        raw_documents = loader.load()

        text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        documents = text_splitter.split_documents(raw_documents)

        extractor = GLiNEREntityExtractor(labels=["person", "topic"])
        documents = extractor.transform_documents(documents)

        print(documents[0].metadata)

    .. code-block:: output

        {'source': 'https://raw.githubusercontent.com/hwchase17/chat-your-data/master/state_of_the_union.txt', 'person': ['president zelenskyy', 'vladimir putin']}

    Parameters
    ----------
    labels : list[str]
        List of entity kinds to extract.
    batch_size : int, default 8
        The number of documents to process in each batch.
    metadata_key_prefix : str, default ""
        A prefix to add to metadata keys outputted by the extractor.
        This will be prepended to the label, with the value (or values) holding the
        generated keywords for that entity kind.
    model : str, default "urchade/gliner_mediumv2.1"
        The GLiNER model to use.

    """  # noqa: E501

    def __init__(
        self,
        labels: list[str],
        *,
        batch_size: int = 8,
        metadata_key_prefix: str = "",
        model: str = "urchade/gliner_mediumv2.1",
    ):
        try:
            from gliner import GLiNER  # type: ignore

            self._model = GLiNER.from_pretrained(model)

        except ImportError:
            raise ImportError(
                "gliner is required for the GLiNEREntityExtractor. "
                "Please install it with `pip install gliner`."
            ) from None

        self._batch_size = batch_size
        self._labels = labels
        self.metadata_key_prefix = metadata_key_prefix

    @override
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        for i in range(0, len(documents), self._batch_size):
            batch = documents[i : i + self._batch_size]
            texts = [item.page_content for item in batch]
            extracted = self._model.batch_predict_entities(
                texts=texts, labels=self._labels, **kwargs
            )
            for i, entities in enumerate(extracted):
                labels = set()
                for entity in entities:
                    label = self.metadata_key_prefix + entity["label"]
                    labels.add(label)
                    batch[i].metadata.setdefault(label, set()).add(
                        entity["text"].lower()
                    )
                for label in labels:
                    batch[i].metadata[label] = list(batch[i].metadata[label])
        return documents



import json

from langchain_core.documents import Document
from langchain_graph_retriever.document_transformers.metadata_denormalizer import (
    DEFAULT_PATH_DELIMITER,
    DEFAULT_STATIC_VALUE,
    DENORMALIZED_KEYS_KEY,
    MetadataDenormalizer,
)


def test_transform_documents(animal_docs: list[Document]):
    original_docs = [animal_docs[0]]
    transformer = MetadataDenormalizer()

    # confirm "keywords" contains a list value in original document
    list_key = "keywords"
    original_metadata = original_docs[0].metadata
    assert isinstance(original_metadata[list_key], list)

    # pull the first keyword in that list
    first_keyword = original_metadata[list_key][0]

    # transform the document
    denormalized_docs = transformer.transform_documents(original_docs)
    denormalized_metadata = denormalized_docs[0].metadata

    # confirm "keywords" no longer exists as a metadata key
    assert list_key not in denormalized_metadata

    # confirm "keywords" exists as a denormalized key
    assert list_key in json.loads(denormalized_metadata[DENORMALIZED_KEYS_KEY])

    # confirm the denormalized key has the expected value
    denormalized_key = f"{list_key}{DEFAULT_PATH_DELIMITER}{first_keyword}"
    assert denormalized_key in denormalized_metadata
    assert denormalized_metadata[denormalized_key] == DEFAULT_STATIC_VALUE


def test_revert_documents(animal_docs: list[Document]):
    original_docs = [animal_docs[0]]
    transformer = MetadataDenormalizer()

    denormalized_docs = transformer.transform_documents(original_docs)

    reverted_docs = transformer.revert_documents(denormalized_docs)

    assert original_docs[0] == reverted_docs[0]



from collections.abc import Sequence
from typing import Any

from langchain_core.documents import BaseDocumentTransformer, Document
from typing_extensions import override


class HierarchyLinkExtractor(BaseDocumentTransformer):
    r"""
    Extract links from a document hierarchy.

    Example
    -------

    .. code-block:: python

        # Given three paths (in this case, within the "Root" document):
        h1 = ["Root", "H1"]
        h1a = ["Root", "H1", "a"]
        h1b = ["Root", "H1", "b"]

        # Parent links `h1a` and `h1b` to `h1`.
        # Child links `h1` to `h1a` and `h1b`.
        # Sibling links `h1a` and `h1b` together (both directions).

    Example use with documents
    --------------------------
    .. code_block: python
        transformer = LinkExtractorTransformer([
            HierarchyLinkExtractor().as_document_extractor(
                # Assumes the "path" to each document is in the metadata.
                # Could split strings, etc.
                lambda doc: doc.metadata.get("path", [])
            )
        ])
        linked = transformer.transform_documents(docs)

    Parameters
    ----------
    path_metadata_key : str, default "path"
        Metadata key containing the path.
        This may correspond to paths in a file system, hierarchy in a document, etc.
    path_delimiter : str, default "\"
        Delimiter of items in the path.
    parent_links : bool, default True
        Whether to link each document to it's parent.
        If `True`, `metadata["parent_out"]` will be populated to link to
        `metadata["parent_in"]`.
    child_links: bool, default False
        Whether to link each document from a section to its children.
        If `True`, `metadata["child_out"]` will be populated to link to
        `metadata["child_in"]`.
    sibling_links : bool, default False
        Whether to link each document to sibling (adjacent) documents.
        If `True`, `metadata["sibling"]` will be populated.
    """

    def __init__(
        self,
        *,
        path_metadata_key: str = "path",
        path_delimiter: str = "\\",
        parent_links: bool = True,
        child_links: bool = False,
        sibling_links: bool = False,
    ):
        self._path_metadata_key = path_metadata_key
        self._path_delimiter = path_delimiter
        self._parent_links = parent_links
        self._child_links = child_links
        self._sibling_links = sibling_links

    @override
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        for document in documents:
            if self._path_metadata_key not in document.metadata:
                msg = (
                    f"path not found in metadata at {self._path_metadata_key}",
                    f" for document id: {document.id}",
                )
                raise ValueError(msg)

            path: str = document.metadata[self._path_metadata_key]
            if self._parent_links:
                # This is linked from everything with this parent path.
                document.metadata["parent_in"] = path
            if self._child_links:
                # This is linked to every child with this as it's "parent" path.
                document.metadata["child_out"] = path

            path_parts = path.split(self._path_delimiter)
            if len(path_parts) >= 1:
                parent_path = self._path_delimiter.join(path_parts[0:-1])
                if self._parent_links and len(path_parts) > 1:
                    # This is linked to the nodes with the given parent path.
                    document.metadata["parent_out"] = parent_path
                if self._child_links and len(path_parts) > 1:
                    # This is linked from every node with the given parent path.
                    document.metadata["child_in"] = parent_path
                if self._sibling_links:
                    # This is a sibling of everything with the same parent.
                    document.metadata["sibling"] = parent_path
        return documents



from collections.abc import Sequence
from typing import Any

import spacy  # type: ignore
from langchain_core.documents import BaseDocumentTransformer, Document
from spacy.language import Language  # type: ignore
from typing_extensions import override


class SpacyNERTransformer(BaseDocumentTransformer):
    """
    Add metadata to documents about named entities using `spaCy`.

    Preliminaries
    -------------

    Install the ``spacy`` package. The below uses ``spacy[apple]`` so it includes optimizations
    for running on Apple M1 hardrware.

    .. code-block:: bash

        pip install -q "spacy[apple]"

    Download the model.

    .. code-block:: bash

        python -m spacy download en_core_web_sm

    Example
    -------
    We load the ``state_of_the_union.txt`` file, chunk it, then for each chunk we
    add named entities to the metadata.

    .. code-block:: python

        from langchain_community.document_loaders import WebBaseLoader
        from langchain_text_splitters import CharacterTextSplitter
        from langchain_graph_retriever.document_transformers import SpacyNERTransformer

        loader = WebBaseLoader(
            "https://raw.githubusercontent.com/hwchase17/chat-your-data/master/state_of_the_union.txt"
        )
        raw_documents = loader.load()

        text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        documents = text_splitter.split_documents(raw_documents)

        extractor = SpacyNERTransformer()
        documents = extractor.transform_documents(documents)

        print(documents[0].metadata)

    .. code-block:: output

        {'source': 'https://raw.githubusercontent.com/hwchase17/chat-your-data/master/state_of_the_union.txt', 'person': ['president zelenskyy', 'vladimir putin']}

    Parameters
    ----------
    include_labels : set[str], optional
        Set of entity labels to include. Will include all labels if empty.
    exclude_labels : set[str], optional
        Set of entity lables to exclude. Will not exclude anything if empty.
    metadata_key : str, default ""
        The metadata key to store the extracted entities in.
    model : str, default "en_core_web_sm"
        The spacy model to use.

    Notes
    -----
     See spaCy docs for the selected model to determine what NER labels will be
     used. The default model
    (en_core_web_sm)[https://spacy.io/models/en#en_core_web_sm-labels] produces:
    CARDINAL, DATE, EVENT, FAC, GPE, LANGUAGE, LAW, LOC, MONEY, NORP, ORDINAL,
    ORG, PERCENT, PERSON, PRODUCT, QUANTITY, TIME, WORK_OF_ART.

    """  # noqa: E501

    def __init__(
        self,
        *,
        include_labels: set[str] = set(),
        exclude_labels: set[str] = set(),
        limit: int | None = None,
        metadata_key: str = "entities",
        model: str | Language = "en_core_web_sm",
    ):
        self.include_labels = include_labels
        self.exclude_labels = exclude_labels
        self.limit = limit
        self.metadata_key = metadata_key

        if isinstance(model, str):
            self.model = spacy.load(model)
        elif isinstance(model, Language):
            self.model = model
        else:
            raise ValueError(f"Invalid model: {model}")

    @override
    def transform_documents(
        self, documents: Sequence[Document], **kwargs: Any
    ) -> Sequence[Document]:
        docs = []
        for doc in documents:
            results = self.model(doc.page_content).ents
            # Filter and de-duplicate entities.
            entities = list(
                {
                    f"{e.label_}: {e.text}"
                    for e in results
                    if not self.include_labels or e.label_ in self.include_labels
                    if not self.exclude_labels or e.label_ not in self.exclude_labels
                }
            )
            # Limit it, if necessary.
            if self.limit:
                entities = entities[: self.limit]
            docs.append(
                Document(
                    id=doc.id,
                    page_content=doc.page_content,
                    metadata={self.metadata_key: entities, **doc.metadata},
                )
            )
        return docs
